@@ -8,13 +8,14 @@
  */
 
 #include <afsconfig.h>
-#include "../afs/param.h"
+#include "afs/param.h"
 
-RCSID("$Header: /cvs/openafs/src/afs/UKERNEL/osi_vfsops.c,v 1.6 2002/03/25 17:11:56 shadow Exp $");
+RCSID
+    ("$Header: /cvs/openafs/src/afs/UKERNEL/osi_vfsops.c,v 1.10 2003/07/15 23:14:29 shadow Exp $");
 
-#include "../afs/sysincludes.h"	/* Standard vendor system headers */
-#include "../afs/afsincludes.h"	/* Afs-based standard headers */
-#include "../afs/afs_stats.h"   /* statistics stuff */
+#include "afs/sysincludes.h"	/* Standard vendor system headers */
+#include "afsincludes.h"	/* Afs-based standard headers */
+#include "afs/afs_stats.h"	/* statistics stuff */
 
 
 int afs_mount();
@@ -36,52 +37,58 @@ struct vcache *afs_globalVp = 0;
 int afs_rootCellIndex = 0;
 
 #if !defined(AFS_USR_AIX_ENV)
-#include "../sys/syscall.h"
+#include "sys/syscall.h"
 #endif
 
 afs_mount(afsp, path, data)
-    char *path;
-    caddr_t data; 
-    struct vfs *afsp;
+     char *path;
+     caddr_t data;
+     struct vfs *afsp;
 {
     AFS_STATCNT(afs_mount);
 
     if (afs_globalVFS) {
 	/* Don't allow remounts since some system (like AIX) don't handle it well */
- 	return (setuerror(EBUSY));
+	return (setuerror(EBUSY));
     }
     afs_globalVFS = afsp;
     afsp->vfs_bsize = 8192;
-    afsp->vfs_fsid.val[0] = AFS_VFSMAGIC; /* magic */
-    afsp->vfs_fsid.val[1] = (afs_int32) AFS_VFSFSID; 
+    afsp->vfs_fsid.val[0] = AFS_VFSMAGIC;	/* magic */
+    afsp->vfs_fsid.val[1] = (afs_int32) AFS_VFSFSID;
 
     return 0;
 }
 
-afs_unmount (afsp)
-    struct vfs *afsp; {
+afs_unmount(afsp)
+     struct vfs *afsp;
+{
     AFS_STATCNT(afs_unmount);
     afs_globalVFS = 0;
     afs_shutdown();
     return 0;
 }
 
-afs_root (OSI_VFS_ARG(afsp), avpp)
+afs_root(OSI_VFS_ARG(afsp), avpp)
     OSI_VFS_DECL(afsp);
-    struct vnode **avpp; {
+     struct vnode **avpp;
+{
     register afs_int32 code = 0;
     struct vrequest treq;
-    register struct vcache *tvp=0;
+    register struct vcache *tvp = 0;
     OSI_VFS_CONVERT(afsp)
 
-    AFS_STATCNT(afs_root);
+	AFS_STATCNT(afs_root);
     if (afs_globalVp && (afs_globalVp->states & CStatd)) {
 	tvp = afs_globalVp;
     } else {
-	if (!(code = afs_InitReq(&treq, u.u_cred)) &&
-	    !(code = afs_CheckInit())) {
-	    tvp = afs_GetVCache(&afs_rootFid, &treq, (afs_int32 *)0,
-				(struct vcache*)0, WRITE_LOCK);
+	if (afs_globalVp) {
+	    afs_PutVCache(afs_globalVp);
+	    afs_globalVp = NULL;
+	}
+
+	if (!(code = afs_InitReq(&treq, u.u_cred))
+	    && !(code = afs_CheckInit())) {
+	    tvp = afs_GetVCache(&afs_rootFid, &treq, NULL, NULL);
 	    /* we really want this to stay around */
 	    if (tvp) {
 		afs_globalVp = tvp;
@@ -92,7 +99,7 @@ afs_root (OSI_VFS_ARG(afsp), avpp)
     if (tvp) {
 	VN_HOLD(AFSTOV(tvp));
 
-	AFSTOV(tvp)->v_flag |= VROOT;	    /* No-op on Ultrix 2.2 */
+	AFSTOV(tvp)->v_flag |= VROOT;	/* No-op on Ultrix 2.2 */
 	afs_globalVFS = afsp;
 	*avpp = AFSTOV(tvp);
     }
@@ -103,32 +110,32 @@ afs_root (OSI_VFS_ARG(afsp), avpp)
 }
 
 afs_sync(afsp)
-    struct vfs *afsp; 
+     struct vfs *afsp;
 {
     AFS_STATCNT(afs_sync);
     return 0;
 }
 
 afs_statfs(afsp, abp)
-    register struct vfs *afsp;
-    struct statfs *abp;
- {
-	AFS_STATCNT(afs_statfs);
-	abp->f_type = 0;
-	abp->f_bsize = afsp->vfs_bsize;
-	abp->f_fsid.val[0] = AFS_VFSMAGIC; /* magic */
-	abp->f_fsid.val[1] = (afs_int32) AFS_VFSFSID;
-	return 0;
+     register struct vfs *afsp;
+     struct statfs *abp;
+{
+    AFS_STATCNT(afs_statfs);
+    abp->f_type = 0;
+    abp->f_bsize = afsp->vfs_bsize;
+    abp->f_fsid.val[0] = AFS_VFSMAGIC;	/* magic */
+    abp->f_fsid.val[1] = (afs_int32) AFS_VFSFSID;
+    return 0;
 }
 
 afs_mountroot()
 {
     AFS_STATCNT(afs_mountroot);
-    return(EINVAL);
+    return (EINVAL);
 }
 
-afs_swapvp() 
+afs_swapvp()
 {
     AFS_STATCNT(afs_swapvp);
-    return(EINVAL);
+    return (EINVAL);
 }
